"""
Either creates an n-quads files containing the content of a Manifest file or uploads the content to Fuseki.

It creates:

 1. A Named Graph for each resource using the item's IRI as the graph IRI
 2. A Named Graph for the catalogue, either using the catalogue's IRI as the graph IRI + "-catalogue" if given, or by making one up - a Blank Node
 3. All the triples in resources with roles mrr:CompleteCatalogueAndResourceLabels & mrr:IncompleteCatalogueAndResourceLabels within a Named Graph with IRI <https://background>
 4. An Olis Virtual Graph, <https://olis.dev/VirtualGraph> object using the catalogue IRI, if give, which is as an alias for all the Named Graphs from 1., 2. & 3.
 5. Multiple entries in the System Graph - Named Graph with IRI <https://olis.dev/SystemGraph> - for each Named and the Virtual Graph from 1., 2. & 3.

Run this script with the -h flag for more help, i.e. ~$ python loader.py -h
"""

import argparse
import sys
from pathlib import Path

from kurra.format import make_dataset, export_quads
from kurra.fuseki import upload
from kurra.utils import load_graph
from rdflib import DCAT, DCTERMS, OWL, PROF, RDF, SDO, SKOS
from rdflib import Graph, URIRef, Dataset
from typing import Literal as TLiteral
import logging

try:
    from prezmanifest import MRR, OLIS, validate, __version__
    from prezmanifest.utils import get_files_from_artifact
except ImportError:
    import sys

    sys.path.append(str(Path(__file__).parent.parent.resolve()))
    from prezmanifest import MRR, OLIS, validate, __version__
    from prezmanifest.utils import get_files_from_artifact


def load(
    manifest: Path,
    sparql_endpoint: str = None,
    destination_file: Path = None,
    return_data_type: TLiteral["Graph", "Dataset", None] = None
) -> None | Graph | Dataset:
    """Loads a catalogue of data from a prezmanifest file, whose content are valid according to the Prez Manifest Model
    (https://kurrawong.github.io/prez.dev/manifest/) either into a specified quads file in the Trig format, or into a
    given SPARQL Endpoint."""

    if return_data_type == "Dataset":
        dataset_holder = Dataset()

    if return_data_type == "Graph":
        graph_holder = Graph()

    return_data_value_error_message = "return_data_type was set to an invalid value. Must be one of Dataset or Graph or None"

    def _export(data: Graph | Dataset, iri, sparql_endpoint, destination_file, return_data_type, append=False):
        if type(data) is Dataset:
            if iri is not None:
                raise ValueError("If the data is a Dataset, the parameter iri must be None")

            if destination_file is not None:
                export_quads(data, destination_file)
            elif sparql_endpoint is not None:
                for g in data.graphs():
                    if g.identifier != URIRef("urn:x-rdflib:default"):
                        _export(g, g.identifier, sparql_endpoint, None, None)
            else:
                if return_data_type == "Dataset":
                    return data
                elif return_data_type == "Graph":
                    gx = Graph()
                    for g in data.graphs():
                        if g.identifier != URIRef("urn:x-rdflib:default"):
                            for s, p, o in g.triples((None, None, None)):
                                gx.add((s, p, o))
                    return gx

        elif type(data) is Graph:
            if iri is None:
                raise ValueError("If the data is a GRaph, the parameter iri must not be None")

            msg = f"exporting {iri} "
            if destination_file is not None:
                msg += f"to file {destination_file} "
                export_quads(make_dataset(data, iri), destination_file)
            elif sparql_endpoint is not None:
                msg += f"to SPARQL Endpoint {sparql_endpoint}"
                upload(sparql_endpoint, data, iri, append)
            else:  # returning data
                if return_data_type == "Dataset":
                    msg += "to Dataset"
                    for s, p, o in data:
                        dataset_holder.add((s, p, o, iri))
                elif return_data_type == "Graph":
                    msg += "to Graph"
                    for s, p, o in data:
                        graph_holder.add((s, p, o))
                else:
                    raise ValueError(return_data_value_error_message)

            logging.info(msg)

    if sum(x is not None for x in [sparql_endpoint, destination_file, return_data_type]) != 1:
        raise ValueError(
            "You must specify exactly 1 of sparql_endpoint, destination_file or return_data_type",
        )

    MANIFEST_ROOT_DIR = manifest.parent
    # load and validate manifest
    validate(manifest)
    manifest_graph = load_graph(manifest)

    vg = Graph()
    vg_iri = None

    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        for role in manifest_graph.objects(o, PROF.hasRole):
            # The catalogue - must be processed first
            if role == MRR.CatalogueData:
                for artifact in manifest_graph.objects(o, PROF.hasArtifact):
                    # load the Catalogue, determine the Virtual Graph & Catalogue IRIs
                    # and fail if we can't see a Catalogue object
                    c = load_graph(MANIFEST_ROOT_DIR / str(artifact))
                    vg_iri = c.value(predicate=RDF.type, object=DCAT.Catalog)
                    if vg_iri is None:
                        raise ValueError(
                            f"ERROR: Could not create a Virtual Graph as no Catalog found in the Catalogue data"
                        )
                    catalogue_iri = URIRef(str(vg_iri) + "-catalogue")

                    # add to the System Graph
                    vg.add((vg_iri, RDF.type, OLIS.VirtualGraph))
                    vg.add((vg_iri, OLIS.isAliasFor, catalogue_iri))
                    vg_name = c.value(
                        subject=vg_iri,
                        predicate=SDO.name | DCTERMS.title | SKOS.prefLabel,
                    ) or str(vg_iri)
                    vg.add((vg_iri, SDO.name, vg_name))

                    # export the Catalogue data
                    _export(c, catalogue_iri, sparql_endpoint, destination_file, return_data_type)

        # non-catalogue resources
        for s, o in manifest_graph.subject_objects(PROF.hasResource):
            for role in manifest_graph.objects(o, PROF.hasRole):
                # The data files & background - must be processed after Catalogue
                if role in [
                    MRR.CompleteCatalogueAndResourceLabels,
                    MRR.IncompleteCatalogueAndResourceLabels,
                    MRR.ResourceData,
                ]:
                    for artifact in manifest_graph.objects(o, PROF.hasArtifact):
                        for f in get_files_from_artifact(manifest, artifact):
                            if str(f.name).endswith(".ttl"):
                                fg = Graph().parse(f)
                                # fg.bind("rdf", RDF)

                                if role == MRR.ResourceData:
                                    resource_iri = fg.value(predicate=RDF.type, object=SKOS.ConceptScheme) or fg.value(predicate=RDF.type, object=OWL.Ontology)

                                if role in [
                                    MRR.CompleteCatalogueAndResourceLabels,
                                    MRR.IncompleteCatalogueAndResourceLabels,
                                ]:
                                    resource_iri = URIRef("http://background")

                                if resource_iri is None:
                                    raise ValueError(f"Could not determine Resource IRI for file {f}")

                                vg.add((vg_iri, OLIS.isAliasFor, resource_iri))

                                # export one Resource
                                _export(fg, resource_iri, sparql_endpoint, destination_file, return_data_type)
                            elif str(f.name).endswith(".trig"):
                                d = Dataset()
                                d.parse(f)
                                for g in d.graphs():
                                    if g.identifier != URIRef("urn:x-rdflib:default"):
                                        vg.add((vg_iri, OLIS.isAliasFor, g.identifier))
                                _export(d, None, sparql_endpoint, destination_file, return_data_type)


        # export the System Graph
        _export(vg, OLIS.SystemGraph, sparql_endpoint, destination_file, return_data_type, append=True)

    if return_data_type == "Dataset":
        return dataset_holder
    elif return_data_type == "Graph":
        return graph_holder
    elif return_data_type is None:
        pass  # return nothing
    else:
        raise ValueError(return_data_value_error_message)


def setup_cli_parser(args=None):
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="{version}".format(version=__version__),
    )

    group.add_argument(
        "-e",
        "--endpoint",
        help="The SPARQL endpoint you want to load the data into. Cannot be specified when destination is.",
    )

    group.add_argument(
        "-d",
        "--destination",
        help="The n-quads file you want to export the data into. Cannot be specified when endpoint is.",
    )

    parser.add_argument(
        "manifest",
        help="A Manifest file to process",
        type=Path,
    )

    return parser.parse_args(args)


def cli(args=None):
    if args is None:
        args = sys.argv[1:]

    args = setup_cli_parser(args)

    load(
        Path(args.manifest),
        args.endpoint if args.endpoint is not None else None,
        Path(args.destination) if args.destination is not None else None,
    )


if __name__ == "__main__":
    retval = cli(sys.argv[1:])
    if retval is not None:
        sys.exit(retval)



"""
This script creates a "Prez Resources" table in either Markdown or ASCIIDOC from a Manifest file which it validates first

Run this script with the -h flag for more help, i.e. ~$ python documentor.py -h

Example:

Input:

PREFIX mrr: <https://prez.dev/ManifestResourceRoles/>
PREFIX prez: <https://prez.dev/>
PREFIX prof: <http://www.w3.org/ns/dx/prof/>
PREFIX schema: <https://schema.org/>

[]
    a prez:Manifest ;
    prof:hasResource
        [
            prof:hasArtifact "catalogue.ttl" ;
            prof:hasRole mrr:ContainerData ;
            schema:description "The definition of, and medata for, the container which here is a dcat:Catalog object" ;
            schema:name "Catalogue Definition"
        ] ,
        [
            prof:hasArtifact "vocabs/*.ttl" ;
            prof:hasRole mrr:ContentData ;
            schema:description "skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder" ;
            schema:name "Content"
        ] ,
        [
            prof:hasArtifact "https://github.com/RDFLib/prez/blob/main/prez/reference_data/profiles/ogc_records_profile.ttl" ;
            prof:hasRole mrr:ContainerAndContentModel ;
            schema:description "The default Prez profile for Records API" ;
            schema:name "Profile Definition"
        ] ,
        [
            prof:hasArtifact "_background/labels.ttl" ;
            prof:hasRole mrr:CompleteContainerAndContentLabels ;
            schema:description "An RDF file containing all the labels for the container content" ;
        ] ;
    .

Output:

Resource | Role | Description
--- | --- | ---
Catalogue Definition, [`catalogue.ttl`](catalogue.ttl) | [Container Data](https://prez.dev/ManifestResourceRoles/ContainerData) | The definition of, and medata for, the container which here is a dcat:Catalog object
Content, [`vocabs/*.ttl`](vocabs/*.ttl) | [Content Data](https://prez.dev/ManifestResourceRoles/ContentData) | skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder
Profile Definition, [`ogc_records_profile.ttl`](https://github.com/RDFLib/prez/blob/main/prez/reference_data/profiles/ogc_records_profile.ttl) | [Container & Content Model](https://prez.dev/ManifestResourceRoles/ContainerAndContentModel) | The default Prez profile for Records API
Labels file, [`_background/labels.ttl`](_background/labels.ttl) | [Complete Content and Container Labels](https://prez.dev/ManifestResourceRoles/CompleteContainerAndContentLabels) | An RDF file containing all the labels for the container content
"""

import argparse
import sys
from pathlib import Path
from urllib.parse import ParseResult, urlparse

from rdflib import DCAT, DCTERMS, PROF, SDO, SKOS, Graph
from kurra.utils import load_graph

try:
    from prezmanifest import MRR, validate, __version__
    from prezmanifest.utils import *
except ImportError:
    import sys

    sys.path.append(str(Path(__file__).parent.parent.resolve()))
    from prezmanifest import MRR, validate, __version__
    from prezmanifest.utils import *


def create_table(manifest: Path, t="markdown") -> str:
    # load and validate manifest
    validate(manifest)
    manifest_graph = load_graph(manifest)

    # add in MRR vocab
    manifest_graph += load_graph(Path(__file__).parent / "mrr.ttl")

    if t == "asciidoc":
        header = "|===\n| Resource | Role | Description\n\n"
    else:
        header = "Resource | Role | Description\n--- | --- | ---\n"

    body = ""
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        a = str(manifest_graph.value(o, PROF.hasArtifact))
        if t == "asciidoc":
            artifact = f'link:{a}[`{a.split("/")[-1] if a.startswith("http") else a}`]'
        else:
            artifact = f'[`{a.split("/")[-1] if a.startswith("http") else a}`]({a})'
        role_iri = manifest_graph.value(o, PROF.hasRole)
        role_label = manifest_graph.value(role_iri, SKOS.prefLabel)
        if t == "asciidoc":
            role = f"{role_iri}[{role_label}]"
        else:
            role = f"[{role_label}]({role_iri})"
        name = manifest_graph.value(o, SDO.name)
        description = manifest_graph.value(o, SDO.description)
        n = f"{name}, {artifact}" if name is not None else f"{artifact}"
        d = description if description is not None else ""
        if t == "asciidoc":
            body += f"| {n} | {role} | {d}\n"
        else:
            body += f"{n} | {role} | {d}\n"

    if t == "asciidoc":
        footer = "|===\n"
    else:
        footer = ""

    return (header + body + footer).strip()


def create_catalogue(manifest: Path):
    MANIFEST_ROOT_DIR = manifest.parent
    # load and validate manifest
    validate(manifest)
    manifest_graph = load_graph(manifest)

    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        for role in manifest_graph.objects(o, PROF.hasRole):
            if role == MRR.CatalogueData:
                for artifact in manifest_graph.objects(o, PROF.hasArtifact):
                    # the artifact can only be a triples file (not a quads file)
                    catalogue = load_graph(MANIFEST_ROOT_DIR / artifact)

    # get the IRI of the catalogue
    catalogue_iri = catalogue.value(predicate=RDF.type, object=DCAT.Catalog)

    # non-catalogue resources
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        for role in manifest_graph.objects(o, PROF.hasRole):
            if role in [
                MRR.CompleteCatalogueAndResourceLabels,
                MRR.IncompleteCatalogueAndResourceLabels,
                MRR.ResourceData,
            ]:
                for artifact in manifest_graph.objects(o, PROF.hasArtifact):
                    for f in get_files_from_artifact(manifest, artifact):
                        for iri in sorted(get_identifier_from_file(f)):
                            if iri != URIRef("urn:x-rdflib:default"):
                                catalogue.add((catalogue_iri, DCTERMS.hasPart, iri))

    return catalogue


def setup_cli_parser(args=None):
    def url_file_or_folder(input: str) -> ParseResult | Path:
        parsed = urlparse(input)
        if all([parsed.scheme, parsed.netloc]):
            return parsed
        path = Path(input)
        if path.is_file():
            return path
        if path.is_dir():
            return path
        raise argparse.ArgumentTypeError(
            f"{input} is not a valid input. Must be a file, folder or sparql endpoint"
        )

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="{version}".format(version=__version__),
    )

    parser.add_argument(
        "function",
        help="The documentation function you wish to perform",
        choices=["table", "catalogue"],
    )

    parser.add_argument(
        "-t",
        "--type",
        help="The type of markup you want to export: Markdown or ASCCIDOC. Only relevant for the 'table' function",
        choices=["markdown", "asciidoc"],
        default="markdown",
    )

    parser.add_argument(
        "input",
        help="A Prez Manifest file",
        type=url_file_or_folder,
    )

    return parser.parse_args(args)


def cli(args=None):
    if args is None:
        args = sys.argv[1:]

    args = setup_cli_parser(args)

    if args.function == "table":
        print(create_table(args.input, t=args.type))
    else:
        print(create_catalogue(args.input).serialize(format="longturtle"))


if __name__ == "__main__":
    retval = cli(sys.argv[1:])
    if retval is not None:
        sys.exit(retval)



from pathlib import Path
from typing import List
from collections.abc import Generator
from rdflib import Literal, URIRef, Graph, Dataset
from rdflib.namespace import OWL, RDF, RDFS, SKOS


def get_files_from_artifact(manifest: Path, artifact: Literal) -> List[Path] | Generator[Path]:
    """Returns an iterable (list or generator) of Path objects for files within an artifact literal.

    This function will correctly interpret artifacts such as 'file.ttl', '*.ttl', '**/*.trig' etc."""
    if not "*" in str(artifact):
        return [manifest.parent / Path(str(artifact))]
    else:
        artifact_str = str(artifact)
        glob_marker_location = artifact_str.find("*")
        glob_parts = [artifact_str[:glob_marker_location], artifact_str[glob_marker_location:]]

        return Path(manifest.parent / Path(glob_parts[0])).glob(glob_parts[1])


def get_identifier_from_file(file: Path) -> List[URIRef]:
    """Returns a list if RDFLib graph identifier (URIRefs) from a triples or quads file
    for all owl:Ontology and skos:ConceptScheme objects"""
    if file.name.endswith(".ttl"):
        g = Graph().parse(file)
        return [
            g.value(predicate=RDF.type, object=SKOS.ConceptScheme) or
            g.value(predicate=RDF.type, object=OWL.Ontology)
        ]
    elif file.name.endswith(".trig"):
        gs = []
        d = Dataset()
        d.parse(file, format="trig")
        for g in d.graphs():
            gs.append(g.identifier)
        return gs
    else:
        return []